#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "cdi.h"
#include "tinecdi.h"
#include "cdiopc.h"
#include "open62541.h"

#define BUSPLUG_NAME "OPCUA"
char gBusPlugName[64] = BUSPLUG_NAME;
void addNodes(int cdiLine, int *addr, int offset, int number);
int getMethods(int cdiLine, int *addr, int devnr, char *bwrName, char *nodeName);
UA_StatusCode getValueFromMonitor(UA_Client *client, const UA_NodeId nodeId, const char *nodeName, UA_Variant *outValue);

#define UA_INTERATE_DEFAULT_TIMEOUT 10
THRHANDLE hUACyclerThread = (THRHANDLE)NULL;
UA_UInt32 UA_interate_to = UA_INTERATE_DEFAULT_TIMEOUT;
THRHANDLE UA_CycleCreateThread(THREAD_PARAM lpvThreadParam);

int cdiInitBusPlug(void)
{
  int cc = 0;

  /* any filter function for data on the bus ? */
  if ((cc = cdiRegisterBusFilter(gBusPlugName, opcuaFilter)) != 0) goto err;
  /* how to get things started:*/
  if ((cc = cdiRegisterBusInitialization(gBusPlugName, opcuaInit)) != 0) goto err;
  /* where the action is: */
  if ((cc = cdiRegisterBusHandler(gBusPlugName, opcuaHandler)) != 0) goto err;
  /* how to clean things up when we're unloaded */
  if ((cc = cdiRegisterBusCleanup(gBusPlugName, opcuaExit)) != 0) goto err;
  /* any diagnostic scanning functions ?  */
  if ((cc = cdiRegisterBusScanner(gBusPlugName, opcuaScanBus)) != 0) goto err;
  /* our cycler !  */
  if ((cc = cdiRegisterCycleFunction(gBusPlugName, 10, opcuaCycle)) != 0) goto err;
err:
  if (cc)
  {
    cdilog("%s : failure in bus registration SEDPC", erlst[cc]);
  }
  return cc;
}
void fixBusPlugName(void)
{
  char *ptr = getenv("CDI_OPCUA_PLUG");
  if (ptr != NULL && strlen(ptr) > 0)
  {
    strncpy(gBusPlugName, ptr, 16);
    gBusPlugName[16] = 0;
  }
}
#if defined(WIN32)
BOOL WINAPI DllMain(HINSTANCE hInstDll, DWORD fdwReason, LPVOID lpvReserved)
{
  switch (fdwReason)
  {
  case DLL_PROCESS_ATTACH:  // DLL being loaded
    fixBusPlugName();
    cdiInitBusPlug();
    break;
  case DLL_THREAD_ATTACH:
    break;
  case DLL_THREAD_DETACH:
    break;
  case DLL_PROCESS_DETACH:  // DLL being unloaded
    break;
  }
  return (TRUE);
}
#elif defined(UNIX)
__attribute__((constructor)) void cdiMain(void)
{
  fixBusPlugName();
  cdiInitBusPlug();
  return;
}
#endif

UA_Client *client = NULL;
char ua_target[256] = "";
CDIOPCUA_EXPORT int opcuaInit(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  UA_StatusCode retval = -1;
  cdi_mutex_t tmpMutex;
  int lastError = 0, error = 0, cc = 0;
  if (parameterList == NULL || strlen(parameterList) == 0) ccerr(ERROR_INVALID_PARAMETER);
  /* try to get Mutex if this is the first application  */
  lastError = cdiMutexInit(&tmpMutex, cdiLine);
  if (lastError == ERROR_ALREADY_EXISTS) cdiMutexLock(&tmpMutex);
  client = UA_Client_new();
  UA_ClientConfig_setDefault(UA_Client_getConfig(client));
  memset(ua_target, 0, 255);
  snprintf(ua_target, 254, "opc.tcp://%s", parameterList);
  retval = UA_Client_connect(client, ua_target);
  if (retval != UA_STATUSCODE_GOOD) 
  {
    UA_Client_delete(client);
    error = (int)retval;
    client = NULL;
  }
# ifdef UA_THREAD
  if (hUACyclerThread != (THRHANDLE)0) goto err;
  if ((hUACyclerThread=UA_CycleCreateThread(NULL)) == (THRHANDLE)0)
  {
    cdilog("could not launch UA Cycler thread");
  }
# endif
err:
  if (lastError == ERROR_ALREADY_EXISTS) cdiMutexUnlock(&tmpMutex);
  return error;
}

UA_DataType cdi2uaType(int fmt)
{
  switch (fmt)
  {
    case cdiByte: return UA_TYPES[UA_TYPES_BYTE];
    case cdiShort: return UA_TYPES[UA_TYPES_INT16];
    case cdiInteger: return UA_TYPES[UA_TYPES_INT32];
    case cdiFloat: return UA_TYPES[UA_TYPES_FLOAT];
    case cdiDouble: return UA_TYPES[UA_TYPES_DOUBLE];
    case cdiChar: return UA_TYPES[UA_TYPES_SBYTE];
    default: return UA_TYPES[UA_TYPES_UINT32];
  }
}
int uaType2cdi(UA_DataType *t)
{
  switch (t->typeIndex)
  {
    case UA_TYPES_BYTE: return cdiByte;
    case UA_TYPES_SBYTE: return cdiChar;
    case UA_TYPES_UINT16:
    case UA_TYPES_INT16: return cdiShort;
    case UA_TYPES_UINT32: 
    case UA_TYPES_INT32: return cdiInteger;
    case UA_TYPES_FLOAT: return cdiFloat;
    case UA_TYPES_DOUBLE: return cdiDouble;
  }
  return cdiDouble;
}
int needToReconnect(UA_StatusCode sts)
{
  switch (sts)
  {
    case UA_STATUSCODE_GOOD: return FALSE;
    case UA_STATUSCODE_BADSERVERNOTCONNECTED:
    case UA_STATUSCODE_BADCONNECTIONCLOSED:
    case UA_STATUSCODE_BADINTERNALERROR:
      return TRUE;
    case UA_STATUSCODE_BADDISCONNECT:
      return TRUE;
    default:
      return FALSE;
  }
  return FALSE;
}

#define MAX_ITEM_KEY_SIZE 256
typedef struct CdiMonitoredItem
{
  UA_NodeId nodeId;
  UA_UInt32 subId;
  UA_UInt32 monId;
  char key[MAX_ITEM_KEY_SIZE]; // void *monContext;
  UA_DataValue value;
  UA_DateTime timestamp;
  UA_StatusCode sts;
  int hasUpdated;
  struct CdiMonitoredItem *nxt;
} CdiMonitoredItem;

CdiMonitoredItem *monItems[LNKTBL_HASH_SIZE] = {NULL};
CdiMonitoredItem *getCdiMonitoredItem(char *key)
{
  CdiMonitoredItem *item;
  int idx;
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);
  for (item=monItems[idx]; item != NULL; item = item->nxt)
  {
    if (strnicmp(item->key, key, FULL_NAME_SIZE)) continue;
    break;
  }
  return item;
}
CdiMonitoredItem *addCdiMonitoredItem(UA_NodeId nodeId, UA_UInt32 subId, UA_UInt32 monId, char *key)
{
  CdiMonitoredItem *item = getCdiMonitoredItem(key);
  int idx;
  if (item != NULL) return item; // already added
  if (key == NULL) return NULL;
  item = (CdiMonitoredItem *)calloc(1, sizeof(CdiMonitoredItem));
  if (item == NULL) return NULL;
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);
  strncpy(item->key, key, MAX_ITEM_KEY_SIZE);
  item->monId = monId;
  item->subId = subId;
  item->nodeId = nodeId;
  item->sts = UA_STATUSCODE_BADNODATA;
  item->nxt = monItems[idx];
  monItems[idx] = item;
  return item;
}
void rmvCdiMonitoredItem(char *key)
{
  CdiMonitoredItem *item = getCdiMonitoredItem(key), *prv = NULL;
  int idx;
  if (item == NULL) return; // not in the list
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);
  for (prv = item = monItems[idx]; item != NULL; prv = item, item = item->nxt)
  {
    if (!strnicmp(item->key, key, FULL_NAME_SIZE))
    { // found it
      prv->nxt = item->nxt;
      if (prv == item) monItems[idx] = item->nxt;
      free(item);
      return;
    }
  }
  return;
}
THREAD_TYPE UA_clnThreadTask(THREAD_PARAM lpvThreadParam)
{
  UNUSED(lpvThreadParam);
  //SetThreadName("UA_ClnCycle");
  cdilog("UA client cycle thread starting");
  while (!ServerExitCondition)
  {
    UA_Client_run_iterate(client, UA_interate_to);
  }
  feclog("UA client cycle thread terminating");
  return 0;
}
#ifdef UNIX
THRHANDLE UA_CycleCreateThread(THREAD_PARAM lpvThreadParam)
{
  pthread_t tid;
  int cc = pthread_create(&tid, gPtrClnCycleThreadAttr, UA_clnThreadTask, lpvThreadParam);
  return cc ? 0 : (THRHANDLE)tid;
}
#else
DWORD dwUAClnThreadId;
THRHANDLE UA_CycleCreateThread(THREAD_PARAM lpvThreadParam)
{
  return (THRHANDLE)CreateThread(NULL, 0, UA_clnThreadTask, lpvThreadParam, 0, &dwUAClnThreadId);
}
#endif

CDIOPCUA_EXPORT void opcuaCycle(void)
{
  UA_Client_run_iterate(client, UA_interate_to);
}

CDIOPCUA_EXPORT void opcuaHandler(CdiRequestInfoBlk *pReq)
{
  UA_StatusCode retval = -1;
  UA_Int16 nsidx = pReq->pDev->cdiAddr[0];
  UA_Variant value; /* Variants can hold scalar values and arrays of any type */
  UA_DataType dtype = cdi2uaType(pReq->pDev->dataFormat);
  int dsize = getTypeLength(pReq->pDev->dataFormat); // in bytes
  int isMonitor=FALSE;
  char *nodeName = pReq->pDev->extName == NULL ? pReq->pDev->lngName : pReq->pDev->extName;
  UA_NodeId nodeId = UA_NODEID_STRING(nsidx, nodeName);

  if (!strnicmp(pReq->pDev->devName, "$DISCOVER", 9))
  {
    char ddev[64];
    strncpy(ddev, &pReq->pDev->devName[9], 32); ddev[32] = 0;
    if (strlen(ddev) > 0)
    {
      char *c = &ddev[1]; /* skip 1st ':' */
      int line = 0, idx = 0, rstart = atoi(c), rstop = rstart + 1;
      if ((c=strchr(c, ':')) != NULL) rstop = atoi(++c); else rstop = rstart, rstart = 0;
      findDevice(pReq->pDev->devName, &line, &idx);
      if (rstart == 0) rstart = 1; // cdi devices start at #1
      addNodes(line,pReq->pDev->cdiAddr,rstart,rstop);
    }
    return;
  }
  //UA_Client_run_iterate(client, UA_interate_to);
  isMonitor = pReq->pDev->cdiAddr[1];
  if (dsize == 0)
  {
    pReq->errorCode = illegal_format;
    return;
  }
  UA_Variant_init(&value);

  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);

  switch (pReq->accessFlag)
  {
    default: /* there are more complicated cases, but let's keep it simple here */
    case cdiReadFlag:
    {
      caseRead:
      dsize *= (pReq->inLength / dsize);
      if (isMonitor)
        retval = getValueFromMonitor(client, nodeId, nodeName, &value);
      else
        retval = UA_Client_readValueAttribute(client, nodeId, &value);
      if (needToReconnect(retval))
      {
        UA_Client_disconnect(client);
        if (UA_Client_connect(client, ua_target) == UA_STATUSCODE_GOOD) goto caseRead;
      }
      if (value.data == NULL) retval = UA_STATUSCODE_BADDECODINGERROR;
      if (retval == UA_STATUSCODE_GOOD)
      {
        memcpy(pReq->pInData, value.data, dsize);
        // pReq->responseTime << there's currently no timestamp variable !
      }
      pReq->errorCode = (retval == UA_STATUSCODE_GOOD) ? 0 : hardware_error;
      break;
    }
    case cdiWriteReadFlag:
    case cdiWriteFlag:
    {
      int nelems = pReq->outLength / dsize;
      if (nelems > 1)
        UA_Variant_setArrayCopy(&value, pReq->pOutData, nelems, &dtype);
      else
        UA_Variant_setScalarCopy(&value, pReq->pOutData, &dtype);
      caseWrite:
      retval = UA_Client_writeValueAttribute(client, nodeId, &value);
      if (needToReconnect(retval))
      {
        if (UA_Client_connect(client, ua_target) == UA_STATUSCODE_GOOD) goto caseWrite;
      }
      pReq->errorCode = (retval == UA_STATUSCODE_GOOD) ? 0 : hardware_error;
      if (pReq->accessFlag == cdiWriteReadFlag) goto caseRead;
      break;
    }
  }
  UA_Variant_clear(&value);

  if (pReq->errorCode != 0)
  { /* bus action not successful */
    if (pReq->perror != NULL) *(pReq->perror) = pReq->errorCode;
    pReq->deviceError = pReq->errorCode;
    pReq->pDev->numberDeviceError++;
  }

  pReq->active = finishRequest; /* must signal that the request is finished */
  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  return;
}

CDIOPCUA_EXPORT int opcuaExit(int ln)
{
  if (client != NULL) UA_Client_delete(client); /* Disconnects the client internally */
  return 0;
}

CDIOPCUA_EXPORT int opcuaFilter(int *parameter, int accessFlag)
{
  /* data can be filter according to the bus's filtering criteria here */
  return 0;
}

CDIOPCUA_EXPORT int opcuaScanBus(char *target, char *result, int resultBufferSize)
{
  return 0;
}

static void opcMonitorCallback(UA_Client *client, UA_UInt32 subId, void *subContext,
  UA_UInt32 monId, void *monContext, UA_DataValue *value)
{
  CdiMonitoredItem *item = getCdiMonitoredItem((char *)monContext);
  UA_Variant v;
  if (item == NULL)
  {
    return;
  }
  item->hasUpdated = TRUE;
  UA_DataValue_copy((const UA_DataValue *)value, &item->value);
  item->sts = value->status;
}

UA_StatusCode getValueFromMonitor(UA_Client *client, const UA_NodeId nodeId, const char *nodeName,
  UA_Variant *outValue)
{
  CdiMonitoredItem *item = getCdiMonitoredItem(nodeName);
  if (item == NULL)
  {
    time_t t0 = time(NULL);
    UA_UInt32 subId = 0;
    UA_UInt32 monId = 0;
    UA_UInt32 mid = startCdiMonitor(&subId, &monId, nodeName);
    item = addCdiMonitoredItem(nodeId, subId, monId, nodeName);
    while (TRUE)
    {
      UA_Client_run_iterate(client, UA_interate_to);
      if (time(NULL) > t0 + 5)
      {
        item->sts = UA_STATUSCODE_BADTIMEOUT;
        item->hasUpdated = TRUE;
      }
      if (item->hasUpdated) break;
    }
  }
  UA_Variant_copy((const UA_Variant *)&item->value.value, outValue);
  item->timestamp = item->value.hasSourceTimestamp ? item->value.sourceTimestamp : item->value.serverTimestamp;
  return item->sts;
}

UA_UInt32 startCdiMonitor(UA_UInt32 *subId, UA_UInt32 *monId,const char *nodeName)
{
  UA_CreateSubscriptionRequest request = UA_CreateSubscriptionRequest_default();
  UA_CreateSubscriptionResponse response = UA_Client_Subscriptions_create(client, request,
    NULL, NULL, NULL);

  if (response.responseHeader.serviceResult == UA_STATUSCODE_GOOD)
  {
    *subId = response.subscriptionId;
  }
  UA_MonitoredItemCreateRequest monRequest =
    UA_MonitoredItemCreateRequest_default(UA_NODEID_STRING(1, nodeName));

  UA_MonitoredItemCreateResult monResponse =
    UA_Client_MonitoredItems_createDataChange(client, response.subscriptionId,
      UA_TIMESTAMPSTORETURN_BOTH,
      monRequest, nodeName, opcMonitorCallback, NULL);
  if (monResponse.statusCode == UA_STATUSCODE_GOOD)
  {
    *monId = monResponse.monitoredItemId;
  }
  return 0;
}

int getMethods(int cdiLine, int *addr, int devnr, char *bwrName, char *nodeName)
{
  int i, k, len, fmt, n = 0, nsIndex;
  char dev[256], lngnam[256];
  UA_BrowseRequest bmReq;
  UA_BrowseResponse bmResp;
  UA_NodeId nodeId;
  UA_Variant value; /* Variants can hold scalar values and arrays of any type */
  UA_StatusCode retval = -1;
  if (addr == NULL) return 0;
  nsIndex = addr[0];
  UA_BrowseRequest_init(&bmReq);
  bmReq.requestedMaxReferencesPerNode = 0;
  bmReq.nodesToBrowse = UA_BrowseDescription_new();
  bmReq.nodesToBrowseSize = 1;
  bmReq.nodesToBrowse[0].nodeId = UA_NODEID_STRING_ALLOC(nsIndex, nodeName);
  bmReq.nodesToBrowse[0].resultMask = UA_BROWSERESULTMASK_ALL; /* return everything */
  bmResp = UA_Client_Service_browse(client, bmReq);

  for (i = 0; i < bmResp.resultsSize; i++)
  {
    for (k = 0; k < bmResp.results[i].referencesSize; k++)
    {
      UA_ReferenceDescription *ref = &(bmResp.results[i].references[k]);
      if (ref->nodeId.nodeId.namespaceIndex != nsIndex) continue;
      if (ref->nodeId.nodeId.identifierType != UA_NODEIDTYPE_STRING) continue;
      memset(dev, 0, 255); memset(lngnam, 0, 255);
      len = MIN((int)ref->browseName.name.length, 255);
      snprintf(dev, 32, "%s.%.*s", bwrName, len, ref->browseName.name.data);
      //strncpy(dev, ref->browseName.name.data, len);
      len = MIN((int)ref->nodeId.nodeId.identifier.string.length, 255);
      strncpy(lngnam, ref->nodeId.nodeId.identifier.string.data, len);
      if (cdi_RegisterDevice(cdiLine, dev, lngnam, "opc desc", -1, devnr, addr, 2) == 0)
      {
        // TODO: how to find the OPCua data type ?
        //int dt = UA_findDataType(&ref->nodeId.nodeId);
        UA_NodeId typeId;
        nodeId = UA_NODEID_STRING(nsIndex, lngnam);
        retval = UA_Client_readValueAttribute(client, nodeId, &value);
        UA_Client_readDataTypeAttribute(client, nodeId, &typeId);
        fmt = uaType2cdi(UA_findDataType(&typeId));
        //fmt = (retval == UA_STATUSCODE_GOOD) ? uaType2cdi(value.type) : cdiDouble;
        cdi_assignDataFormat(dev, fmt);
        cdi_assignDefaultInterval(dev, 500);
        n++; // count them 
        devnr++; // increment the current device number 
      }
    }
  }

  UA_BrowseRequest_clear(&bmReq);  //<<< BUG in OPCua code ...
  UA_BrowseResponse_clear(&bmResp);
  return n;
}
void addNodes(int cdiLine, int *addr, int offset, int number)
{
  int i, k, len, nsIndex;
  char dev[256], lngnam[256];
  UA_BrowseRequest bReq;
  UA_BrowseResponse bResp;
  if (addr == NULL) return;
  nsIndex = addr[0];
  UA_BrowseRequest_init(&bReq);
  bReq.requestedMaxReferencesPerNode = 0;
  bReq.nodesToBrowse = UA_BrowseDescription_new();
  bReq.nodesToBrowseSize = 1;
  bReq.nodesToBrowse[0].nodeId = UA_NODEID_NUMERIC(0, UA_NS0ID_OBJECTSFOLDER); /* browse objects folder */
  bReq.nodesToBrowse[0].resultMask = UA_BROWSERESULTMASK_ALL; /* return everything */
  bResp = UA_Client_Service_browse(client, bReq);
  for (i=0; i < bResp.resultsSize; i++) 
  {
    for (k=0; k < bResp.results[i].referencesSize; k++) 
    {
      UA_ReferenceDescription *ref = &(bResp.results[i].references[k]);
      if (ref->nodeId.nodeId.namespaceIndex != nsIndex) continue;
      if (ref->nodeId.nodeId.identifierType != UA_NODEIDTYPE_STRING) continue;
      memset(dev, 0, 255); memset(lngnam, 0, 255);
      len = MIN((int)ref->browseName.name.length, 255);
      strncpy(dev, ref->browseName.name.data, len);
      len = MIN((int)ref->nodeId.nodeId.identifier.string.length, 255);
      strncpy(lngnam, ref->nodeId.nodeId.identifier.string.data, len); 
      offset += getMethods(cdiLine, addr, offset, dev, lngnam);
    }
  }

  UA_BrowseRequest_clear(&bReq);
  UA_BrowseResponse_clear(&bResp);

}
